@@ -43,10 +43,10 @@ typedef unsigned char * AclBits;
  * How a user joins a group.
  * Here is the node to construct one list per user.
  */
-struct usergroup
+struct aclusergroup
 {
-  struct user *u;	/* the user who borrows us his rights */
-  struct usergroup *next;
+  struct acluser *u;	/* the user who borrows us his rights */
+  struct aclusergroup *next;
 };
 #endif /* MULTIUSER */
 
@@ -59,9 +59,9 @@ struct usergroup
  * users is the User entry of the session owner (creator)
  * and anchors all other users. Add/Delete users there.
  */
-typedef struct user
+typedef struct acluser
 {
-  struct user *u_next;		/* continue the main user list */
+  struct acluser *u_next;		/* continue the main user list */
   char u_name[20+1];		/* login name how he showed up */
   char *u_password;		/* his password (may be NullStr). */
   int  u_checkpassword;		/* nonzero if this u_password is valid */
@@ -75,7 +75,7 @@ typedef struct user
 #ifdef MULTIUSER
   int u_id;			/* a uniq index in the bitfields. */
   AclBits u_umask_w_bits[ACL_BITS_PER_WIN];	/* his window create umask */
-  struct usergroup *u_group;	/* linked list of pointers to other users */
+  struct aclusergroup *u_group;	/* linked list of pointers to other users */
 #endif
 } User;
 
